@@ -3,7 +3,7 @@
  *  BlueZ - Bluetooth protocol stack for Linux
  *
  *  Copyright (C) 2006-2007  Nokia Corporation
- *  Copyright (C) 2004-2008  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2004-2009  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -62,9 +62,9 @@
 #include "../src/adapter.h"
 #include "../src/device.h"
 
-#define DC_TIMEOUT 3000
+#define DC_TIMEOUT 3
 
-#define RING_INTERVAL 3000
+#define RING_INTERVAL 3
 
 #define BUF_SIZE 1024
 
@@ -145,6 +145,8 @@ struct headset {
 	gboolean cme_enabled;
 	gboolean cwa_enabled;
 	gboolean pending_ring;
+	gboolean nrec;
+	gboolean nrec_req;
 
 	headset_state_t state;
 	struct pending_connect *pending;
@@ -313,7 +315,7 @@ static int supported_features(struct audio_device *device, const char *buf)
 
 	print_hf_features(hs->hf_features);
 
-	err = headset_send(hs, "\r\n+BRSF:%u\r\n", ag.features);
+	err = headset_send(hs, "\r\n+BRSF: %u\r\n", ag.features);
 	if (err < 0)
 		return err;
 
@@ -325,7 +327,7 @@ static char *indicator_ranges(const struct indicator *indicators)
 	int i;
 	GString *gstr;
 
-	gstr = g_string_new("\r\n+CIND:");
+	gstr = g_string_new("\r\n+CIND: ");
 
 	for (i = 0; indicators[i].desc != NULL; i++) {
 		if (i == 0)
@@ -348,7 +350,7 @@ static char *indicator_values(const struct indicator *indicators)
 	int i;
 	GString *gstr;
 
-	gstr = g_string_new("\r\n+CIND:");
+	gstr = g_string_new("\r\n+CIND: ");
 
 	for (i = 0; indicators[i].desc != NULL; i++) {
 		if (i == 0)
@@ -530,7 +532,7 @@ static void sco_connect_cb(GIOChannel *chan, int err, const bdaddr_t *src,
 		return;
 	}
 
-	debug("MYSCO:SCO socket opened for headset %s", dev->path);
+	debug("SCO socket opened for headset %s", dev->path);
 
 	sk = g_io_channel_unix_get_fd(chan);
 
@@ -551,7 +553,8 @@ static void sco_connect_cb(GIOChannel *chan, int err, const bdaddr_t *src,
 
 	if (hs->pending_ring) {
 		ring_timer_cb(NULL);
-		ag.ring_timer = g_timeout_add(RING_INTERVAL, ring_timer_cb,
+		ag.ring_timer = g_timeout_add_seconds(RING_INTERVAL,
+						ring_timer_cb,
 						NULL);
 		hs->pending_ring = FALSE;
 	}
@@ -629,7 +632,7 @@ static int telephony_generic_rsp(struct audio_device *device, cme_error_t err)
 
 	if (err != CME_ERROR_NONE) {
 		if (hs->cme_enabled)
-			return headset_send(hs, "\r\n+CME ERROR:%d\r\n", err);
+			return headset_send(hs, "\r\n+CME ERROR: %d\r\n", err);
 		else
 			return headset_send(hs, "\r\nERROR\r\n");
 	}
@@ -708,7 +711,7 @@ static int call_hold(struct audio_device *dev, const char *buf)
 		return 0;
 	}
 
-	err = headset_send(hs, "\r\n+CHLD:(%s)\r\n", ag.chld);
+	err = headset_send(hs, "\r\n+CHLD: (%s)\r\n", ag.chld);
 	if (err < 0)
 		return err;
 
@@ -723,6 +726,7 @@ static int call_hold(struct audio_device *dev, const char *buf)
 
 	return 0;
 }
+
 static int button_press(struct audio_device *device, const char *buf)
 {
 	struct headset *hs = device->headset;
@@ -744,25 +748,8 @@ int telephony_answer_call_rsp(void *telephony_device, cme_error_t err)
 	return telephony_generic_rsp(telephony_device, err);
 }
 
-static gboolean hs_open_sco_channel_to_hf(void *data)
-{
-	struct audio_device *device = data;
-	int err;
-	char hs_address[18];
-
-	err = sco_connect(device, NULL, NULL, NULL);
-	if (err < 0) 
-		return FALSE;
-	ba2str(&device->dst, hs_address);
-	info("Sco Connection opened from %s, %s", hs_address, device->path);
-	return TRUE;
-}
-
 static int answer_call(struct audio_device *device, const char *buf)
 {
-	gboolean success = TRUE;
-	struct headset *hs = device->headset;
-
 	if (ag.ring_timer) {
 		g_source_remove(ag.ring_timer);
 		ag.ring_timer = 0;
@@ -773,16 +760,7 @@ static int answer_call(struct audio_device *device, const char *buf)
 		ag.number = NULL;
 	}
 
-	if(!hs->sco)
-		success = hs_open_sco_channel_to_hf((void *)device);
-		
-	if(success)
-	{
-		telephony_answer_call_req(device);
-		g_dbus_emit_signal(device->conn, device->path,
-			AUDIO_HEADSET_INTERFACE, "IncomingCallAccepted",
-			DBUS_TYPE_INVALID);
-	}
+	telephony_answer_call_req(device);
 
 	return 0;
 }
@@ -850,7 +828,7 @@ static int response_and_hold(struct audio_device *device, const char *buf)
 	}
 
 	if (ag.rh >= 0)
-		headset_send(hs, "\r\n+BTRH:%d\r\n", ag.rh);
+		headset_send(hs, "\r\n+BTRH: %d\r\n", ag.rh);
 
 	return headset_send(hs, "\r\nOK\r\n", ag.rh);
 }
@@ -862,16 +840,9 @@ int telephony_last_dialed_number_rsp(void *telephony_device, cme_error_t err)
 
 static int last_dialed_number(struct audio_device *device, const char *buf)
 {
-	/* Check if the Last Dial Memory Information is Correct Else Return Error */
-	if(telephony_last_dialed_number_req(device))
-        {
-		g_dbus_emit_signal(device->conn, device->path,
-			AUDIO_HEADSET_INTERFACE, "RedialRequested",
-			DBUS_TYPE_INVALID);
-		return 0;
-	}
-	else 
-	        return -EINVAL;
+	telephony_last_dialed_number_req(device);
+
+	return 0;
 }
 
 int telephony_dial_number_rsp(void *telephony_device, cme_error_t err)
@@ -894,11 +865,9 @@ static int dial_number(struct audio_device *device, const char *buf)
 	memset(number, 0, sizeof(number));
 	strncpy(number, &buf[3], buf_len - 4);
 
-	/* Check for Validity of Last Dial Information */
-	if(!telephony_dial_number_req(device, number))
-		return -EINVAL;
-	return 0;
+	telephony_dial_number_req(device, number);
 
+	return 0;
 }
 
 static int signal_gain_setting(struct audio_device *device, const char *buf)
@@ -1040,12 +1009,23 @@ int telephony_call_hold_rsp(void *telephony_device, cme_error_t err)
 	return telephony_generic_rsp(telephony_device, err);
 }
 
+int telephony_nr_and_ec_rsp(void *telephony_device, cme_error_t err)
+{
+	struct audio_device *device = telephony_device;
+	struct headset *hs = device->headset;
+
+	if (err == CME_ERROR_NONE)
+		hs->nrec = hs->nrec_req;
+
+	return telephony_generic_rsp(telephony_device, err);
+}
+
 int telephony_operator_selection_ind(int mode, const char *oper)
 {
 	if (!active_devices)
 		return -ENODEV;
 
-	send_foreach_headset(active_devices, hfp_cmp, "\r\n+COPS:%d,0,%s\r\n",
+	send_foreach_headset(active_devices, hfp_cmp, "\r\n+COPS: %d,0,%s\r\n",
 				mode, oper);
 	return 0;
 }
@@ -1070,6 +1050,23 @@ static int operator_selection(struct audio_device *device, const char *buf)
 	return 0;
 }
 
+static int nr_and_ec(struct audio_device *device, const char *buf)
+{
+	struct headset *hs = device->headset;
+
+	if (strlen(buf) < 9)
+		return -EINVAL;
+
+	if (buf[8] == '0')
+		hs->nrec_req = FALSE;
+	else
+		hs->nrec_req = TRUE;
+
+	telephony_nr_and_ec_req(device, hs->nrec_req);
+
+	return 0;
+}
+
 static struct event event_callbacks[] = {
 	{ "ATA", answer_call },
 	{ "ATD", dial_number },
@@ -1089,6 +1086,7 @@ static struct event event_callbacks[] = {
 	{ "AT+CMEE", extended_errors },
 	{ "AT+CCWA", call_waiting_notify },
 	{ "AT+COPS", operator_selection },
+	{ "AT+NREC", nr_and_ec },
 	{ 0 }
 };
 
@@ -1453,46 +1451,6 @@ static DBusMessage *hs_is_playing(DBusConnection *conn, DBusMessage *msg,
 	return reply;
 }
 
-static DBusMessage *hs_answer_call(DBusConnection *conn, DBusMessage *msg,
-
-					void *data)
-{
-	DBusMessage *reply = NULL;
-
-	reply = dbus_message_new_method_return(msg);
-	if (!reply)
-		return NULL;
-	
-	if(hs_open_sco_channel_to_hf(data))
-	{
-        	telephony_answer_outgoingcall_req(active_devices->data);
-	}
-	return reply;
-}
-
-static DBusMessage *hs_close_audio(DBusConnection *conn, DBusMessage *msg,
-					void *data)
-{
-	struct audio_device *device = data;
-	struct headset *hs = device->headset;
-	DBusMessage *reply = NULL;
-	char hs_address[18];
-
-	reply = dbus_message_new_method_return(msg);
-	if (!reply)
-		return NULL;
-
-	if (hs->state == HEADSET_STATE_DISCONNECTED)
-		return g_dbus_create_error(msg, ERROR_INTERFACE
-						".NotConnected",
-						"Device not Connected");
-
-	headset_set_state(device, HEADSET_STATE_CONNECTED);
-	ba2str(&device->dst, hs_address);
-	info("Sco Connection Closed from %s, %s", hs_address, device->path);
-	return reply;
-}
-
 static DBusMessage *hs_disconnect(DBusConnection *conn, DBusMessage *msg,
 					void *data)
 {
@@ -1603,7 +1561,8 @@ static DBusMessage *hs_ring(DBusConnection *conn, DBusMessage *msg,
 	}
 
 	ring_timer_cb(NULL);
-	ag.ring_timer = g_timeout_add(RING_INTERVAL, ring_timer_cb, NULL);
+	ag.ring_timer = g_timeout_add_seconds(RING_INTERVAL, ring_timer_cb,
+						NULL);
 
 done:
 	return reply;
@@ -1897,39 +1856,10 @@ static DBusMessage *hs_set_property(DBusConnection *conn,
 
 	return invalid_args(msg);
 }
-
-static DBusMessage *hs_open_audio(DBusConnection *conn, DBusMessage *msg,
-					void *data)
-{
-	struct audio_device *device = data;
-	DBusMessage *reply = NULL;
-	struct headset *hs = device->headset;
-
-
-	reply = dbus_message_new_method_return(msg);
-
-	if (hs->state == HEADSET_STATE_DISCONNECTED )
-		return g_dbus_create_error(msg, ERROR_INTERFACE
-						".NotConnected",
-						"Device not Connected");
-	if (!reply)
-		return NULL;
-	if(!hs_open_sco_channel_to_hf(data))
-	{
-		debug("hs_open_audio:Sco Connection Failed From AG");
-	}
-
-	return reply;
-}
-
 static GDBusMethodTable headset_methods[] = {
 	{ "Connect",		"",	"",	hs_connect,
 						G_DBUS_METHOD_FLAG_ASYNC },
 	{ "Disconnect",		"",	"",	hs_disconnect },
-	{ "CloseSco",		"",	"",	hs_close_audio },
-	{ "OpenSco",		"",	"",	hs_open_audio },
-	{ "AnswerCall",		"",	"",	hs_answer_call,
-						G_DBUS_METHOD_FLAG_ASYNC },
 	{ "IsConnected",	"",	"b",	hs_is_connected },
 	{ "IndicateCall",	"",	"",	hs_ring },
 	{ "CancelCall",		"",	"",	hs_cancel_call },
@@ -1960,8 +1890,6 @@ static GDBusSignalTable headset_signals[] = {
 	{ "SpeakerGainChanged",		"q",	G_DBUS_SIGNAL_FLAG_DEPRECATED },
 	{ "MicrophoneGainChanged",	"q",	G_DBUS_SIGNAL_FLAG_DEPRECATED },
 	{ "CallTerminated",		""	},
-	{ "IncomingCallAccepted",		""	},
-	{ "RedialRequested",		""	},
 	{ "PropertyChanged",		"sv"	},
 	{ NULL, NULL }
 };
@@ -2093,6 +2021,7 @@ struct headset *headset_init(struct audio_device *dev, uint16_t svc,
 	hs->search_hfp = server_is_enabled(&dev->src, HANDSFREE_SVCLASS_ID);
 	hs->hfp_active = FALSE;
 	hs->cli_active = FALSE;
+	hs->nrec = TRUE;
 
 	record = btd_device_get_record(dev->btd_dev, uuidstr);
 	if (!record)
@@ -2192,7 +2121,7 @@ gboolean headset_cancel_stream(struct audio_device *dev, unsigned int id)
 
 	if (hs->auto_dc) {
 		if (hs->rfcomm)
-			hs->dc_timer = g_timeout_add(DC_TIMEOUT,
+			hs->dc_timer = g_timeout_add_seconds(DC_TIMEOUT,
 						(GSourceFunc) hs_dc_timeout,
 						dev);
 		else
@@ -2208,7 +2137,9 @@ static gboolean dummy_connect_complete(struct audio_device *dev)
 	return FALSE;
 }
 
-unsigned int headset_request_stream(struct audio_device *dev, headset_stream_cb_t cb,
+unsigned int headset_request_stream(struct audio_device *dev,
+					headset_stream_cb_t cb,
+					headset_lock_t lock,
 					void *user_data)
 {
 	struct headset *hs = dev->headset;
@@ -2225,7 +2156,8 @@ unsigned int headset_request_stream(struct audio_device *dev, headset_stream_cb_
 		hs->dc_timer = 0;
 	}
 
-	if (hs->state == HEADSET_STATE_CONNECT_IN_PROGRESS)
+	if (hs->state == HEADSET_STATE_CONNECT_IN_PROGRESS ||
+			hs->state == HEADSET_STATE_PLAY_IN_PROGRESS)
 		return connect_cb_new(hs, HEADSET_STATE_PLAYING, cb, user_data);
 
 	if (hs->rfcomm == NULL) {
@@ -2242,6 +2174,66 @@ unsigned int headset_request_stream(struct audio_device *dev, headset_stream_cb_
 	return id;
 }
 
+unsigned int headset_config_stream(struct audio_device *dev,
+					headset_stream_cb_t cb,
+					headset_lock_t lock,
+					void *user_data)
+{
+	struct headset *hs = dev->headset;
+	unsigned int id;
+
+	if (hs->lock & lock)
+		return 0;
+
+	if (hs->dc_timer) {
+		g_source_remove(hs->dc_timer);
+		hs->dc_timer = 0;
+	}
+
+	if (hs->state == HEADSET_STATE_CONNECT_IN_PROGRESS)
+		return connect_cb_new(hs, HEADSET_STATE_CONNECTED, cb,
+					user_data);
+
+	if (hs->rfcomm)
+		goto done;
+
+	if (rfcomm_connect(dev, cb, user_data, &id) < 0)
+		return 0;
+
+	hs->auto_dc = TRUE;
+	hs->pending->target_state = HEADSET_STATE_CONNECTED;
+
+	return id;
+
+done:
+	id = connect_cb_new(hs, HEADSET_STATE_CONNECTED, cb, user_data);
+	g_idle_add((GSourceFunc) dummy_connect_complete, dev);
+	return id;
+}
+
+unsigned int headset_suspend_stream(struct audio_device *dev,
+					headset_stream_cb_t cb,
+					headset_lock_t lock,
+					void *user_data)
+{
+	struct headset *hs = dev->headset;
+	unsigned int id;
+
+	if (hs->lock & ~lock || !hs->rfcomm || !hs->sco)
+		return 0;
+
+	if (hs->dc_timer) {
+		g_source_remove(hs->dc_timer);
+		hs->dc_timer = 0;
+	}
+
+	close_sco(dev);
+
+	id = connect_cb_new(hs, HEADSET_STATE_CONNECTED, cb, user_data);
+	g_idle_add((GSourceFunc) dummy_connect_complete, dev);
+	return id;
+}
+
 gboolean get_hfp_active(struct audio_device *dev)
 {
 	struct headset *hs = dev->headset;
@@ -2276,7 +2268,8 @@ int headset_connect_sco(struct audio_device *dev, GIOChannel *io)
 
 	if (hs->pending_ring) {
 		ring_timer_cb(NULL);
-		ag.ring_timer = g_timeout_add(RING_INTERVAL, ring_timer_cb,
+		ag.ring_timer = g_timeout_add_seconds(RING_INTERVAL,
+						ring_timer_cb,
 						NULL);
 		hs->pending_ring = FALSE;
 	}
@@ -2299,6 +2292,8 @@ static int headset_close_rfcomm(struct audio_device *dev)
 	hs->data_start = 0;
 	hs->data_length = 0;
 
+	hs->nrec = TRUE;
+
 	return 0;
 }
 
@@ -2338,8 +2333,6 @@ void headset_set_state(struct audio_device *dev, headset_state_t state)
 
 	switch (state) {
 	case HEADSET_STATE_DISCONNECTED:
-		/* Ring Is Heard in Most TestCases */
-		telephony_calling_stopped_ind();
 		value = FALSE;
 		close_sco(dev);
 		headset_close_rfcomm(dev);
@@ -2461,7 +2454,7 @@ gboolean headset_unlock(struct audio_device *dev, headset_lock_t lock)
 
 	if (hs->auto_dc) {
 		if (hs->state == HEADSET_STATE_CONNECTED)
-			hs->dc_timer = g_timeout_add(DC_TIMEOUT,
+			hs->dc_timer = g_timeout_add_seconds(DC_TIMEOUT,
 						(GSourceFunc) hs_dc_timeout,
 						dev);
 		else
@@ -2491,6 +2484,13 @@ int headset_get_sco_fd(struct audio_device *dev)
 	return g_io_channel_unix_get_fd(hs->sco);
 }
 
+gboolean headset_get_nrec(struct audio_device *dev)
+{
+	struct headset *hs = dev->headset;
+
+	return hs->nrec;
+}
+
 int telephony_event_ind(int index)
 {
 	if (!active_devices)
@@ -2502,7 +2502,7 @@ int telephony_event_ind(int index)
 	}
 
 	send_foreach_headset(active_devices, hfp_cmp,
-				"\r\n+CIEV:%d,%d\r\n", index + 1,
+				"\r\n+CIEV: %d,%d\r\n", index + 1,
 				ag.indicators[index].val);
 
 	return 0;
@@ -2519,7 +2519,7 @@ int telephony_response_and_hold_ind(int rh)
 	if (ag.rh < 0)
 		return 0;
 
-	send_foreach_headset(active_devices, hfp_cmp, "\r\n+BTRH:%d\r\n",
+	send_foreach_headset(active_devices, hfp_cmp, "\r\n+BTRH: %d\r\n",
 				ag.rh);
 
 	return 0;
@@ -2562,7 +2562,8 @@ int telephony_incoming_call_ind(const char *number, int type)
 	}
 
 	ring_timer_cb(NULL);
-	ag.ring_timer = g_timeout_add(RING_INTERVAL, ring_timer_cb, NULL);
+	ag.ring_timer = g_timeout_add_seconds(RING_INTERVAL, ring_timer_cb,
+						NULL);
 
 	return 0;
 }
@@ -2616,12 +2617,12 @@ int telephony_list_current_call_ind(int idx, int dir, int status, int mode,
 
 	if (number)
 		send_foreach_headset(active_devices, hfp_cmp,
-					"\r\n+CLCC:%d,%d,%d,%d,%d,%s,%d\r\n",
+					"\r\n+CLCC: %d,%d,%d,%d,%d,%s,%d\r\n",
 					idx, dir, status, mode, mprty,
 					number, type);
 	else
 		send_foreach_headset(active_devices, hfp_cmp,
-					"\r\n+CLCC:%d,%d,%d,%d,%d\r\n",
+					"\r\n+CLCC: %d,%d,%d,%d,%d\r\n",
 					idx, dir, status, mode, mprty);
 
 	return 0;
@@ -2633,7 +2634,7 @@ int telephony_subscriber_number_ind(const char *number, int type, int service)
 		return -ENODEV;
 
 	send_foreach_headset(active_devices, hfp_cmp,
-				"\r\n+CNUM:,%s,%d,,%d\r\n",
+				"\r\n+CNUM: ,%s,%d,,%d\r\n",
 				number, type, service);
 
 	return 0;
@@ -2655,7 +2656,7 @@ int telephony_call_waiting_ind(const char *number, int type)
 	if (!active_devices)
 		return -ENODEV;
 
-	send_foreach_headset(active_devices, cwa_cmp, "\r\n+CCWA:%s,%d\r\n",
+	send_foreach_headset(active_devices, cwa_cmp, "\r\n+CCWA: %s,%d\r\n",
 				number, type);
 
 	return 0;
