@@ -314,6 +314,13 @@ static void headset_resume_complete(struct audio_device *dev, void *user_data)
 		goto failed;
 	}
 
+	client->data_fd = headset_get_sco_fd(dev);
+	if (client->data_fd < 0) {
+		error("Unable to get a SCO fd");
+		headset_unlock(client->dev, hs->lock);
+		goto failed;
+	}
+
 	memset(buf, 0, sizeof(buf));
 	rsp->h.type = BT_RESPONSE;
 	rsp->h.name = BT_START_STREAM;
@@ -328,8 +335,6 @@ static void headset_resume_complete(struct audio_device *dev, void *user_data)
 
 	unix_ipc_sendmsg(client, &ind->h);
 
-	client->data_fd = headset_get_sco_fd(dev);
-
 	if (unix_sendmsg_fd(client->sock, client->data_fd) < 0) {
 		error("unix_sendmsg_fd: %s(%d)", strerror(errno), errno);
 		headset_unlock(client->dev, hs->lock);
@@ -919,6 +924,7 @@ static int handle_a2dp_transport(struct unix_client *client,
 	if (client->caps) {
 		g_slist_foreach(client->caps, (GFunc) g_free, NULL);
 		g_slist_free(client->caps);
+		client->caps = NULL;
 	}
 
 	media_transport = avdtp_service_cap_new(AVDTP_MEDIA_TRANSPORT,
